with Unchecked_Deallocation;

package body Conjuntos is
	-- Operaciones auxiliares

	-- Libera sirve para liberar la memoria ocupada por los elementos
	procedure Libera is new Unchecked_Deallocation(Nodo,P_Nodo);    

	-- Copia emplea un esquema de recorrido para realizar una copia
	-- de una lista encadenada, devolviendo un puntero a la nueva lista
	function Copia_Elementos(P: in P_Nodo) return P_Nodo is
		Actual_P, Prim_Copia, Ult_Copia, Nuevo : P_Nodo := null;
	begin
		Actual_P := P;
		while Actual_P /= null loop
			Nuevo := new Nodo'(Actual_P.Elemento,null);
			if Prim_Copia = null then
				Prim_Copia := Nuevo;
			else
				Ult_Copia.Sig := Nuevo;
			end if;
			Ult_Copia := Nuevo;
			Actual_P := Actual_P.Sig;
		end loop;
		return Prim_Copia;
	end Copia_Elementos;

	-- Adjust realiza una copia de la estructura de representacin de un
	-- conjunto asignado para evitar la creacin de un alias
	procedure Adjust(C: in out Conjunto) is
	BEGIN
		C.Elementos  := Copia_Elementos(C.Elementos);
	end Adjust;

	-- Finalize emplea un esquema de recorrido para Liberar la lista de
	-- elementos de la estructura de representacin de un conjunto
	procedure Finalize(C: in out Conjunto) is
		P_Aux1, P_Aux2 : P_Nodo;
	begin
		P_Aux1 := C.Elementos;
		while P_Aux1 /= null loop
			P_Aux2 := P_Aux1;
			P_Aux1 := P_Aux1.Sig;
			Libera(P_Aux2);
		end loop;
		C.NumElementos := 0;
	end Finalize;

	-- Vaco crea y devuelve un conjunto vaco
	function Vaco return Conjunto is
		C: Conjunto;
	begin
		return C; -- C se inicializa como vaco al crearse!!
	end Vaco;

	-- Unitario crea y devuelve un conjunto con un nico elemento (E)
	function Unitario(E: in TElemento) return Conjunto is
		C: Conjunto;
	begin
		C.Elementos    := new Nodo'(E,null);
		C.NumElementos := 1;
		return C;
	end Unitario;

	-- La unin de dos conjuntos se realiza asignando el primero a un 
	-- conjunto auxiliar en el que luego se insertan, mediante un recorrido 
	-- del segundo, todos los elementos de aqul que no pertenecen
	-- al primero.
	function "+"(C1,C2: in Conjunto) return Conjunto is
		C_Aux: Conjunto := C1;
		P_Aux: P_Nodo   := C2.Elementos;
	begin
		-- Se recorre la lista de elementos de C2 para insertar en C_Aux los
		-- que no pertenezcan a C1
		while P_Aux /= null loop
			if not(Pertenece(C1,P_Aux.Elemento)) then
				C_Aux.Elementos   := new Nodo'(P_Aux.Elemento,C_Aux.Elementos);
				C_Aux.NumElementos:= C_Aux.NumElementos + 1;
			end if;
			P_Aux := P_Aux.Sig;
		end loop;

		return C_Aux;
	end "+";

	-- La interseccin de dos conjuntos se realiza insertando en otro 
	-- conjunto los elementos que pertenezcan a los dos pasados por 
	-- parmetros.
	function "*"(C1,C2: in Conjunto) return Conjunto is
	BEGIN
		return C1; -- para que compile
	end "*";

	-- La diferencia se realiza recorriendo el primer conjunto e insertando 
	-- en otro los elementos que no pertenezcan al segundo
	function "-"(C1,C2: in Conjunto) return Conjunto is
	BEGIN
		return C1; -- para que compile
	end "-";

	-- La igualdad se evala mirando si los dos conjuntos tienen el mismo 
	-- tamao y uno est incluido en el otro, o son el mismo
	function "="(C1,C2: in Conjunto) return Boolean is
	begin
		return (Cardinal(C1) = Cardinal(C2)) and then
				 ((C1.Elementos = C2.Elementos) or else
				  (Incluido(C1,C2)));
	end "=";

	-- Incluido hace un recorrido de los elementos de C1 para ver
	-- si todos estn en C2
	function Incluido(C1,C2: in Conjunto) return Boolean is
		Aux: P_Nodo := C1.Elementos;
	begin	
		if C2 /= Vaco then	-- El Conjunto vaco slo contiene al Vaco
			while (Aux /= null) loop
				if not Pertenece(C2,Aux.Elemento) then
	 		    	return False;
				end if;
	 			Aux := Aux.Sig;
			end loop;
			return True;
		else
			Return C1 = Vaco;
		end if;
	end Incluido;
	-- Pertenece busca E entre los elementos de C
	function Pertenece(C: in Conjunto; E: in TElemento) return Boolean is
	BEGIN
		return False; -- para que compile
	end Pertenece;

	-- Cardinal devuelve el nmero de elementos almacenado
	function Cardinal(C: in Conjunto) return Natural is
	begin
		return 0; -- para que compile
	end Cardinal;
end Conjuntos;
