with Ada.Finalization;
use  Ada.Finalization;

generic
	type TElemento is private;
package Conjuntos is
	type Conjunto is private;

	-- Constructoras
	function Vaco return Conjunto;
	-- Representa el conjunto vaco
	function Unitario(E: in TElemento) return Conjunto;
	-- Devuelve un conjuto cuyo nico elemento es E
	function "+"(C1,C2: in Conjunto) return Conjunto;
	-- Devuelve la unin de C1 y C2
	function "*"(C1,C2: in Conjunto) return Conjunto;
	-- Devuelve la interseccin de C1 y C2
	function "-"(C1,C2: in Conjunto) return Conjunto;    
	-- Devuelve la diferencia entre C1 y C2

	-- Observadoras
	function "="(C1,C2: in Conjunto) return Boolean;
	--Devuelve True si y slo si C1 y C2 representan el mismo conjunto
	function Incluido(C1,C2: in Conjunto) return Boolean;
	-- Devuelve True si y slo si el conjunto C1 est contenido en el C2
	function Pertenece(C: in Conjunto; E: in TElemento) return Boolean;
	-- Devuelve True si y slo si el elemento E pertenece al conjunto C 
	function Cardinal(C: in Conjunto) return Natural;
	-- Devuelve el nmero de elementos de C

private
	-- Estructura de representacin del tad Conjunto
	type Nodo;
	type P_Nodo is access Nodo;
	type Nodo is record
		Elemento: TElemento;
		Sig     : P_Nodo;
	end record;

	type Conjunto is new Controlled with record
		Elementos   : P_Nodo := null;
		NumElementos: Natural:= 0;
	end record;

	-- Procedimientos automticos para el control del almacenamiento
	procedure Adjust(c: in out Conjunto);
	procedure Finalize(c: in out Conjunto);
end Conjuntos;
