package conjuntos;
/**
 * La clase Conjunto define objetos que representan un conjunto de objetos.
 */
public class Conjunto implements Cloneable {
	/*
	 * Los objetos de la clase Nodo son unidades de una lista encadenada.
	 */
	private class Nodo implements Cloneable {
		public Object info; // Informacin almacenada en el nodo
		public Nodo siguiente = null; //Siguiente nodo de la lista

		/*
		 * Crea un nodo sin sucesores.
		 */
		public Nodo(Object e) {
			info = e;
		}

		/*
		 * Crea un nodo al principio de una lista de nodos.
		 */
		public Nodo(Object e, Nodo s) {
			info = e;
			siguiente = s;
		}
	}

	private Nodo elementos = null;
	private int numElementos = 0;

	/*
	 * Aade un elemento al principio de la lista de elementos
	 */
	private void aade(Object e) {
		elementos = new Nodo(e, elementos);
		numElementos++;
	}

	private Nodo copiaElementos(Nodo n) {
		if (n == null) {
			return null;
		} else {
			return new Nodo(n.info, copiaElementos(n.siguiente));
		}
	}

	/*
	 * Inicializa el conjunto vaco
	 */
	public Conjunto() {
	}

	/*
	 * Inicializa un conjunto unitario que contiene el objeto e.
	 */
	public Conjunto(Object e) {
		elementos = new Nodo(e);
		numElementos = 1;
	}

	/*
	 * El mtodo clone realiza una copia del conjunto, pero no de
	 * los elementos contenidos en el mismo.
	 */
	public Object clone() throws CloneNotSupportedException {
		Conjunto aux = (Conjunto) super.clone();
		// Se clona la estructura del conjunto
		// Se copia la lista de nodos
		aux.elementos = copiaElementos(elementos);
		return aux;
	}

	/*
	 * La unin de dos conjuntos se realiza copiando el actual en
	 * un conjunto auxiliar en el que luego se insertan, mediante un
	 * recorrido del pasado por parmetro todos los elementos de ste
	 * que no pertenecen al actual.
	 */
	public Conjunto unin(Conjunto c) {
		Conjunto cAux = new Conjunto();

		// Copia del conjunto actual
		cAux.elementos = copiaElementos(elementos);
		cAux.numElementos = numElementos;

		// Adicin de los elementos de c no contenidos
		Nodo nAux = c.elementos;
		while (nAux != null) {
			if (!pertenece(nAux.info)) {
				cAux.aade(nAux.info);
				nAux = nAux.siguiente;
			}
		}
		return cAux;
	}

	/*
	 * La interseccin de dos conjuntos se realiza insertando en otro 
	 * conjunto los elementos que pertenezcan al actual y al pasados
	 * como parmetro. 
	 */
	public Conjunto interseccin(Conjunto c) {
		//Falta desarrollar la operacin (se deja como ejercicio)
		return null;
	}

	/*
	 * La diferencia se realiza recorriendo el conjunto actual e insertando
	 * en otro los elementos que no pertenezcan al pasado como parmetro.
	 */
	public Conjunto diferencia(Conjunto c) {
		//Falta desarrollar la operacin (se deja como ejercicio)
		return null;
	}

	/* 
	 * La igualdad se evala mirando si el actual y c tienen el mismo
	 * tamao y las mismas listas de elementos, o c est incluido
	 * en this. 
	 */
	public boolean equals(Conjunto c) {
		return (this == c) || ((cardinal() == c.cardinal()) && (incluido(c)));
	}

	/*
	 * Incluido recorre la lista de elementos de c, para ver si todos
	 * estn en el actual.
	 */
	public boolean incluido(Conjunto c) {
		if (numElementos < c.numElementos) {
			// Un conjunto no puede estar incluido en otro menor.
			return false;
		} else {
			// Se comprueba que todos los elementos de c estn en el actual.
			Nodo nAux = c.elementos;
			while (nAux != null) {
				if (!pertenece(nAux.info)) {
					return false;
				}
				nAux = nAux.siguiente;
			}
			return true;
		}
	}

	/*
	 * Busca el objeto e entre los elementos del conjunto, devolviendo
	 * true si lo encuentra y false si no.
	 */
	public boolean pertenece(Object e) {
		Nodo nAux = elementos;
		while ((nAux != null) && (!nAux.info.equals(e))) {
			nAux = nAux.siguiente;
		}
		return (nAux != null);
	}

	/*
	 * Cardinal devuelve el nmero de elementos almacenado.
	 */
	public int cardinal() {
		return numElementos;
	}
}
